/* 3D Time domain finite difference kernel
	By: Chris Leader, Abdullah AlTheyab
	Ref: similar to Paulius code but for 2D, then re-extended to 3D
 */
#define BLOCKZ_SIZE 16
#define BLOCKX_SIZE 16
#define BLOCKY_SIZE 16
#define FAT 4
#define COEFFS_SIZE 13

#define C0  0
#define CZ1 1
#define CX1 2
#define CY1 3
#define CZ2 4
#define CX2 5
#define CY2 6
#define CZ3 7
#define CX3 8
#define CY3 9
#define CZ4 10
#define CX4 11
#define CY4 12

#define C_C00(d) (8.0/(5.0*(d)*(d)))
#define get_coeffs(d1,d2,d3) {-1025.0/576.0*(C_C00(d1)+C_C00(d2)+C_C00(d3)), C_C00(d1), C_C00(d2), C_C00(d3),-C_C00(d1)/8.0,-C_C00(d2)/8.0,-C_C00(d3)/8.0,C_C00(d1)/63.0,C_C00(d2)/63.0,C_C00(d3)/63.0,-C_C00(d1)/896.0,-C_C00(d2)/896.0,-C_C00(d3)/896.0}
__constant__ float coeffs[COEFFS_SIZE];

//first off, going to try and use 3D texture
extern "C" __global__ void wave_kernel(float *p0, float *p1, float *p2){
	__shared__ float p1s[BLOCKX_SIZE+2*FAT][BLOCKY_SIZE+2*FAT];

	// define coordinates for the slow axis - everything else will follow from here
	int nblocksz=(n1gpu-2*FAT)/BLOCKZ_SIZE;
	int yblock=(blockIdx.x/nblocksz);
	int y0=FAT+yblock*BLOCKY_SIZE;

	int il=FAT+threadIdx.x; 				// thread-local z coordinate
	int jl=FAT+threadIdx.y;					// thread-local x coordinate

	int ig=(blockIdx.x-yblock*nblocksz)*blockDim.x+il; 	// global z coordinates
	int jg=blockIdx.y*blockDim.y+jl; 	 		// global x coordinates

	float 	p1y_p1, p1y_m1, p1y_p2, p1y_m2, p1y_p3, p1y_m3, p1y_p4, p1y_m4; //here we define the y axis values

	int stride=n1gpu*n2gpu; //this corresponds to distance in elements between wavefield slices
	int addr=ig+n1gpu*jg+stride*y0;
	int addr_fwd=addr-stride-stride-stride-stride; //these (supposedly!) give the memorry location for the current point
		//note that Paulius has a slightly different term here - check one day!

	//copy to shared memory
	p1y_m3=p1[addr_fwd];
	p1y_m2=p1[addr_fwd+=stride];
	p1y_m1=p1[addr_fwd+=stride];
	p1s[jl][il]=p1[addr_fwd+=stride];
	p1y_p1=p1[addr_fwd+=stride];
	p1y_p2=p1[addr_fwd+=stride];
	p1y_p3=p1[addr_fwd+=stride];
	p1y_p4=p1[addr_fwd+=stride];

	// loop over slow axis. ++yl means variable is used before incremented
	for(int yl=0; yl<BLOCKY_SIZE; ++yl){

		int y=y0+yl;

		//update slow axis values
		p1y_m4=p1y_m3;		
		p1y_m3=p1y_m2;		
		p1y_m2=p1y_m1;
		p1y_m1=p1s[jl][il];	
		p1s[jl][il]=p1y_p1;
		p1y_p1=p1y_p2;		
		p1y_p2=p1y_p3;		
		p1y_p3=p1y_p4;
		p1y_p4=p1[addr_fwd+=stride];

    
		if(threadIdx.x<FAT){
			p1s[jl][threadIdx.x]=p1[addr-FAT];
			p1s[jl][il+BLOCKZ_SIZE]=p1[addr+BLOCKZ_SIZE];
		}
		if(threadIdx.y<FAT){
			p1s[threadIdx.y][il]=p1[addr-FAT*n1gpu];
			p1s[jl+BLOCKX_SIZE][il]=p1[addr+BLOCKX_SIZE*n1gpu];
		}
		__syncthreads();

		p2[addr]=//10;
		tex3D(vel_ref1,ig,jg,y)*
			(
			 coeffs[C0]*p1s[jl][il]
			+coeffs[CZ1]*(p1s[jl][il+1]+p1s[jl][il-1])
			+coeffs[CX1]*(p1s[jl+1][il]+p1s[jl-1][il])
			+coeffs[CY1]*(p1y_p1+p1y_m1)
			+coeffs[CZ2]*(p1s[jl][il+2]+p1s[jl][il-2])
			+coeffs[CX2]*(p1s[jl+2][il]+p1s[jl-2][il])
			+coeffs[CY2]*(p1y_p2+p1y_m2)
			+coeffs[CZ3]*(p1s[jl][il+3]+p1s[jl][il-3])
			+coeffs[CX3]*(p1s[jl+3][il]+p1s[jl-3][il])
			+coeffs[CY3]*(p1y_p3+p1y_m3)
			+coeffs[CZ4]*(p1s[jl][il+4]+p1s[jl][il-4])
			+coeffs[CX4]*(p1s[jl+4][il]+p1s[jl-4][il])
			+coeffs[CY4]*(p1y_p4+p1y_m4)
			)
			+p1s[jl][il]+p1s[jl][il]-p0[addr];
			addr+=stride;
	}
}

//same as wave_kernel but uses vel_ref2 (non-random boundary velocity)


extern "C" __global__ void wave_kernel2(float *p0, float *p1, float *p2){
	__shared__ float p1s[BLOCKX_SIZE+2*FAT][BLOCKY_SIZE+2*FAT];

	// define coordinates for the slow axis - everything else will follow from here
	int nblocksz=(n1gpu-2*FAT)/BLOCKZ_SIZE;
	int yblock=(blockIdx.x/nblocksz);
	int y0=FAT+yblock*BLOCKY_SIZE;

	int il=FAT+threadIdx.x; 				// thread-local z coordinate
	int jl=FAT+threadIdx.y;					// thread-local x coordinate

	int ig=(blockIdx.x-yblock*nblocksz)*blockDim.x+il; 	// global z coordinates
	int jg=blockIdx.y*blockDim.y+jl; 	 		// global x coordinates

	float 	p1y_p1, p1y_m1, p1y_p2, p1y_m2, p1y_p3, p1y_m3, p1y_p4, p1y_m4; //here we define the y axis values

	int stride=n1gpu*n2gpu; //this corresponds to distance in elements between wavefield slices
	int addr=ig+n1gpu*jg+stride*y0;
	int addr_fwd=addr-stride-stride-stride-stride; //these (supposedly!) give the memorry location for the current point
		//note that Paulius has a slightly different term here - check one day!

	//copy to shared memory
	p1y_m3=p1[addr_fwd];
	p1y_m2=p1[addr_fwd+=stride];
	p1y_m1=p1[addr_fwd+=stride];
	p1s[jl][il]=p1[addr_fwd+=stride];
	p1y_p1=p1[addr_fwd+=stride];
	p1y_p2=p1[addr_fwd+=stride];
	p1y_p3=p1[addr_fwd+=stride];
	p1y_p4=p1[addr_fwd+=stride];

	// loop over slow axis. ++yl means variable is used before incremented
	for(int yl=0; yl<BLOCKY_SIZE; ++yl){

		int y=y0+yl;

		//update slow axis values
		p1y_m4=p1y_m3;		
		p1y_m3=p1y_m2;		
		p1y_m2=p1y_m1;
		p1y_m1=p1s[jl][il];	
		p1s[jl][il]=p1y_p1;
		p1y_p1=p1y_p2;		
		p1y_p2=p1y_p3;		
		p1y_p3=p1y_p4;
		p1y_p4=p1[addr_fwd+=stride];

    
		if(threadIdx.x<FAT){
			p1s[jl][threadIdx.x]=p1[addr-FAT];
			p1s[jl][il+BLOCKZ_SIZE]=p1[addr+BLOCKZ_SIZE];
		}
		if(threadIdx.y<FAT){
			p1s[threadIdx.y][il]=p1[addr-FAT*n1gpu];
			p1s[jl+BLOCKX_SIZE][il]=p1[addr+BLOCKX_SIZE*n1gpu];
		}
		__syncthreads();

		p2[addr]=
		tex3D(vel_ref2,ig,jg,y)*(
			 coeffs[C0]*p1s[jl][il]
			+coeffs[CZ1]*(p1s[jl][il+1]+p1s[jl][il-1])
			+coeffs[CX1]*(p1s[jl+1][il]+p1s[jl-1][il])
			+coeffs[CY1]*(p1y_p1+p1y_m1)
			+coeffs[CZ2]*(p1s[jl][il+2]+p1s[jl][il-2])
			+coeffs[CX2]*(p1s[jl+2][il]+p1s[jl-2][il])
			+coeffs[CY2]*(p1y_p2+p1y_m2)
			+coeffs[CZ3]*(p1s[jl][il+3]+p1s[jl][il-3])
			+coeffs[CX3]*(p1s[jl+3][il]+p1s[jl-3][il])
			+coeffs[CY3]*(p1y_p3+p1y_m3)
			+coeffs[CZ4]*(p1s[jl][il+4]+p1s[jl][il-4])
			+coeffs[CX4]*(p1s[jl+4][il]+p1s[jl-4][il])
			+coeffs[CY4]*(p1y_p4+p1y_m4)
			)
			+p1s[jl][il]+p1s[jl][il]-p0[addr];
			addr+=stride;
	}

}

extern "C" __global__ void damp_kernel(int nblocksz, float *p0, float *p1){
	int yblock=(blockIdx.x/nblocksz);
	int y0=FAT+yblock*BLOCKY_SIZE;
	int zl=FAT+threadIdx.x; 				// local  zdir
	int zg=(blockIdx.x-yblock*nblocksz)*blockDim.x+zl; 	// global zdir
	int xl=FAT+threadIdx.y; 				// local  xdir
	int xg=blockIdx.y*blockDim.y+xl; 			// global xdir	

	int stride=n1gpu*n2gpu;
	int bc_agpu=50;
	float bc_bgpu=0.0005;

	for(int yl=0; yl<BLOCKY_SIZE; ++yl){
		int y=y0+yl;
		int edge=min2(zg-FAT , min2(y-FAT,min2(min2(n2gpu-FAT-xg,n1gpu-FAT-zg),min2(n3gpu-FAT-y,xg-FAT))));
		if(edge>=0){
		  int addr=zg+n1gpu*xg+stride*y;
		  float temp=expf(-bc_bgpu*(bc_agpu-edge));
		  if(temp<1.){
		    p1[addr]*=temp;
		    p0[addr]*=temp;
		  }
		}
	}
}

extern "C" __global__ void damp_top_kernel(int nblocksz, float *p0, float *p1){
	int yblock=(blockIdx.x/nblocksz);
	int y0=FAT+yblock*BLOCKY_SIZE;
	int zl=FAT+threadIdx.x; 				// local  zdir
	int zg=(blockIdx.x-yblock*nblocksz)*blockDim.x+zl; 	// global zdir
	int xl=FAT+threadIdx.y; 				// local  xdir
	int xg=blockIdx.y*blockDim.y+xl; 			// global xdir	

	int stride=n1gpu*n2gpu;
	int bc_agpu=50;
	float bc_bgpu=0.0005;

	for(int yl=0; yl<BLOCKY_SIZE; ++yl){
		int y=y0+yl;
		int edge=n1gpu-FAT-zg;//min2(zg-FAT ,n1gpu-FAT-zg);
		if(edge>=0){
		  int addr=zg+n1gpu*xg+stride*y;
		  float temp=expf(-bc_bgpu*(bc_agpu-edge));
		  if(temp<1.){
		    p1[addr]*=temp;
		    p0[addr]*=temp;
		  }
		}
	}
}

extern "C" __global__ void wave_kernel_adj(float *p0, float *p1, float *p2){
	__shared__ float p1s[BLOCKX_SIZE+2*FAT][BLOCKY_SIZE+2*FAT];
	__shared__ float vls[BLOCKX_SIZE+2*FAT][BLOCKY_SIZE+2*FAT];

	// define coordinates for the slow axis - everything else will follow from here
	int nblocksz=(n1gpu-2*FAT)/BLOCKZ_SIZE;
	int yblock=(blockIdx.x/nblocksz);
	int y0=FAT+yblock*BLOCKY_SIZE;

	int il=FAT+threadIdx.x; 				// thread-local z coordinate
	int jl=FAT+threadIdx.y;					// thread-local x coordinate

	int ig=(blockIdx.x-yblock*nblocksz)*blockDim.x+il; 	// global z coordinates
	int jg=blockIdx.y*blockDim.y+jl; 	 		// global x coordinates

	float 	p1y_p1, p1y_m1, p1y_p2, p1y_m2, p1y_p3, p1y_m3, p1y_p4, p1y_m4; //here we define the y axis values
	float 	vy_p1, vy_m1, vy_p2, vy_m2, vy_p3, vy_m3, vy_p4, vy_m4; //here we define the y axis values for velocity

	int stride=n1gpu*n2gpu; //this corresponds to distance in elements between wavefield slices
	int addr=ig+n1gpu*jg+stride*y0;
	int addr_fwd=addr-stride-stride-stride-stride; //these (supposedly!) give the memorry location for the current point
		//note that Paulius has a slightly different term here - check one day!

	//copy to shared memory
	p1y_m3=p1[addr_fwd];
	p1y_m2=p1[addr_fwd+=stride];
	p1y_m1=p1[addr_fwd+=stride];
	p1s[jl][il]=p1[addr_fwd+=stride];
	p1y_p1=p1[addr_fwd+=stride];
	p1y_p2=p1[addr_fwd+=stride];
	p1y_p3=p1[addr_fwd+=stride];
	p1y_p4=p1[addr_fwd+=stride];

	vy_m4=tex3D(vel_ref1,ig,jg,y0-4);
	vy_m3=tex3D(vel_ref1,ig,jg,y0-4);
	vy_m2=tex3D(vel_ref1,ig,jg,y0-3);
	vy_m1=tex3D(vel_ref1,ig,jg,y0-2);
	vls[jl][il]=tex3D(vel_ref1,ig,jg,y0-1);
	vy_p1=tex3D(vel_ref1,ig,jg,y0);
	vy_p2=tex3D(vel_ref1,ig,jg,y0+1);
	vy_p3=tex3D(vel_ref1,ig,jg,y0+2);
	vy_p4=tex3D(vel_ref1,ig,jg,y0+3);

	// loop over slow axis. ++yl means variable is used before incremented
	for(int yl=0; yl<BLOCKY_SIZE; ++yl){

		int y=y0+yl;

		//update slow axis values
		p1y_m4=p1y_m3;		
		p1y_m3=p1y_m2;		
		p1y_m2=p1y_m1;
		p1y_m1=p1s[jl][il];	
		p1s[jl][il]=p1y_p1;
		p1y_p1=p1y_p2;		
		p1y_p2=p1y_p3;		
		p1y_p3=p1y_p4;
		p1y_p4=p1[addr_fwd+=stride];

		vy_m4=vy_m3;		
		vy_m3=vy_m2;		
		vy_m2=vy_m1;
		vy_m1=vls[jl][il];	
		vls[jl][il]=vy_p1;
		vy_p1=vy_p2;		
		vy_p2=vy_p3;		
		vy_p3=vy_p4;
		vy_p4=tex3D(vel_ref1,ig,jg,y+4);

		if(threadIdx.x<FAT){
			p1s[jl][threadIdx.x]=p1[addr-FAT];
			p1s[jl][il+BLOCKZ_SIZE]=p1[addr+BLOCKZ_SIZE];
			vls[jl][threadIdx.x]=tex3D(vel_ref1,ig-FAT,jg,y);
			vls[jl][il+BLOCKZ_SIZE]=tex3D(vel_ref1,ig+BLOCKZ_SIZE,jg,y);
		}
		if(threadIdx.y<FAT){
			p1s[threadIdx.y][il]=p1[addr-FAT*n1gpu];
			p1s[jl+BLOCKX_SIZE][il]=p1[addr+BLOCKX_SIZE*n1gpu];
			vls[threadIdx.y][il]=tex3D(vel_ref1,ig,jg-FAT,y);
			vls[jl+BLOCKX_SIZE][il]=tex3D(vel_ref1,ig,jg+BLOCKX_SIZE,y);
		}
		__syncthreads();

		p2[addr]=
			 coeffs[C0]*p1s[jl][il]*vls[jl][il]
			+coeffs[CZ1]*(p1s[jl][il+1]*vls[jl][il+1]+p1s[jl][il-1]*vls[jl][il-1])
			+coeffs[CX1]*(p1s[jl+1][il]*vls[jl+1][il]+p1s[jl-1][il]*vls[jl-1][il])
//			+coeffs[CY1]*(p1y_p1*tex3D(vel_ref1,ig,jg,y+1)+p1y_m1*tex3D(vel_ref1,ig,jg,y-1))
			+coeffs[CY1]*(p1y_p1*vy_p1+p1y_m1*vy_m1)
			+coeffs[CZ2]*(p1s[jl][il+2]*vls[jl][il+2]+p1s[jl][il-2]*vls[jl][il-2])
			+coeffs[CX2]*(p1s[jl+2][il]*vls[jl+2][il]+p1s[jl-2][il]*vls[jl-2][il])
			+coeffs[CY2]*(p1y_p2*vy_p2+p1y_m2*vy_m2)
//			+coeffs[CY2]*(p1y_p2*tex3D(vel_ref1,ig,jg,y+2)+p1y_m2*tex3D(vel_ref1,ig,jg,y-2))
			+coeffs[CZ3]*(p1s[jl][il+3]*vls[jl][il+3]+p1s[jl][il-3]*vls[jl][il-3])
			+coeffs[CX3]*(p1s[jl+3][il]*vls[jl+3][il]+p1s[jl-3][il]*vls[jl-3][il])
			+coeffs[CY3]*(p1y_p3*vy_p3+p1y_m3*vy_m3)
//			+coeffs[CY3]*(p1y_p3*tex3D(vel_ref1,ig,jg,y+3)+p1y_m3*tex3D(vel_ref1,ig,jg,y-3))
			+coeffs[CZ4]*(p1s[jl][il+4]*vls[jl][il+4]+p1s[jl][il-4]*vls[jl][il-4])
			+coeffs[CX4]*(p1s[jl+4][il]*vls[jl+4][il]+p1s[jl-4][il]*vls[jl-4][il])
			+coeffs[CY4]*(p1y_p4*vy_p4+p1y_m4*vy_m4)
//			+coeffs[CY4]*(p1y_p4*tex3D(vel_ref1,ig,jg,y+4)+p1y_m4*tex3D(vel_ref1,ig,jg,y-4))
			+p1s[jl][il]+p1s[jl][il]-p0[addr];
		addr+=stride;
	}

}

extern "C" __global__ void wave_kernel_adj2(float *p0, float *p1, float *p2){
	__shared__ float p1s[BLOCKX_SIZE+2*FAT][BLOCKY_SIZE+2*FAT];
	__shared__ float vls[BLOCKX_SIZE+2*FAT][BLOCKY_SIZE+2*FAT];

	// define coordinates for the slow axis - everything else will follow from here
	int nblocksz=(n1gpu-2*FAT)/BLOCKZ_SIZE;
	int yblock=(blockIdx.x/nblocksz);
	int y0=FAT+yblock*BLOCKY_SIZE;

	int il=FAT+threadIdx.x; 				// thread-local z coordinate
	int jl=FAT+threadIdx.y;					// thread-local x coordinate

	int ig=(blockIdx.x-yblock*nblocksz)*blockDim.x+il; 	// global z coordinates
	int jg=blockIdx.y*blockDim.y+jl; 	 		// global x coordinates

	float 	p1y_p1, p1y_m1, p1y_p2, p1y_m2, p1y_p3, p1y_m3, p1y_p4, p1y_m4; //here we define the y axis values
	float 	vy_p1, vy_m1, vy_p2, vy_m2, vy_p3, vy_m3, vy_p4, vy_m4; //here we define the y axis values for velocity

	int stride=n1gpu*n2gpu; //this corresponds to distance in elements between wavefield slices
	int addr=ig+n1gpu*jg+stride*y0;
	int addr_fwd=addr-stride-stride-stride-stride; //these (supposedly!) give the memorry location for the current point
		//note that Paulius has a slightly different term here - check one day!

	//copy to shared memory
	p1y_m3=p1[addr_fwd];
	p1y_m2=p1[addr_fwd+=stride];
	p1y_m1=p1[addr_fwd+=stride];
	p1s[jl][il]=p1[addr_fwd+=stride];
	p1y_p1=p1[addr_fwd+=stride];
	p1y_p2=p1[addr_fwd+=stride];
	p1y_p3=p1[addr_fwd+=stride];
	p1y_p4=p1[addr_fwd+=stride];

	vy_m4=tex3D(vel_ref2,ig,jg,y0-4);
	vy_m3=tex3D(vel_ref2,ig,jg,y0-4);
	vy_m2=tex3D(vel_ref2,ig,jg,y0-3);
	vy_m1=tex3D(vel_ref2,ig,jg,y0-2);
	vls[jl][il]=tex3D(vel_ref2,ig,jg,y0-1);
	vy_p1=tex3D(vel_ref2,ig,jg,y0);
	vy_p2=tex3D(vel_ref2,ig,jg,y0+1);
	vy_p3=tex3D(vel_ref2,ig,jg,y0+2);
	vy_p4=tex3D(vel_ref2,ig,jg,y0+3);

	// loop over slow axis. ++yl means variable is used before incremented
	for(int yl=0; yl<BLOCKY_SIZE; ++yl){

		int y=y0+yl;

		//update slow axis values
		p1y_m4=p1y_m3;		
		p1y_m3=p1y_m2;		
		p1y_m2=p1y_m1;
		p1y_m1=p1s[jl][il];	
		p1s[jl][il]=p1y_p1;
		p1y_p1=p1y_p2;		
		p1y_p2=p1y_p3;		
		p1y_p3=p1y_p4;
		p1y_p4=p1[addr_fwd+=stride];

		vy_m4=vy_m3;		
		vy_m3=vy_m2;		
		vy_m2=vy_m1;
		vy_m1=vls[jl][il];	
		vls[jl][il]=vy_p1;
		vy_p1=vy_p2;		
		vy_p2=vy_p3;		
		vy_p3=vy_p4;
		vy_p4=tex3D(vel_ref2,ig,jg,y+4);

		if(threadIdx.x<FAT){
			p1s[jl][threadIdx.x]=p1[addr-FAT];
			p1s[jl][il+BLOCKZ_SIZE]=p1[addr+BLOCKZ_SIZE];
			vls[jl][threadIdx.x]=tex3D(vel_ref2,ig-FAT,jg,y);
			vls[jl][il+BLOCKZ_SIZE]=tex3D(vel_ref2,ig+BLOCKZ_SIZE,jg,y);
		}
		if(threadIdx.y<FAT){
			p1s[threadIdx.y][il]=p1[addr-FAT*n1gpu];
			p1s[jl+BLOCKX_SIZE][il]=p1[addr+BLOCKX_SIZE*n1gpu];
			vls[threadIdx.y][il]=tex3D(vel_ref2,ig,jg-FAT,y);
			vls[jl+BLOCKX_SIZE][il]=tex3D(vel_ref2,ig,jg+BLOCKX_SIZE,y);
		}
		__syncthreads();

		p2[addr]=
			 coeffs[C0]*p1s[jl][il]*vls[jl][il]
			+coeffs[CZ1]*(p1s[jl][il+1]*vls[jl][il+1]+p1s[jl][il-1]*vls[jl][il-1])
			+coeffs[CX1]*(p1s[jl+1][il]*vls[jl+1][il]+p1s[jl-1][il]*vls[jl-1][il])
//			+coeffs[CY1]*(p1y_p1*tex3D(vel_ref2,ig,jg,y+1)+p1y_m1*tex3D(vel_ref2,ig,jg,y-1))
			+coeffs[CY1]*(p1y_p1*vy_p1+p1y_m1*vy_m1)
			+coeffs[CZ2]*(p1s[jl][il+2]*vls[jl][il+2]+p1s[jl][il-2]*vls[jl][il-2])
			+coeffs[CX2]*(p1s[jl+2][il]*vls[jl+2][il]+p1s[jl-2][il]*vls[jl-2][il])
			+coeffs[CY2]*(p1y_p2*vy_p2+p1y_m2*vy_m2)
//			+coeffs[CY2]*(p1y_p2*tex3D(vel_ref2,ig,jg,y+2)+p1y_m2*tex3D(vel_ref2,ig,jg,y-2))
			+coeffs[CZ3]*(p1s[jl][il+3]*vls[jl][il+3]+p1s[jl][il-3]*vls[jl][il-3])
			+coeffs[CX3]*(p1s[jl+3][il]*vls[jl+3][il]+p1s[jl-3][il]*vls[jl-3][il])
			+coeffs[CY3]*(p1y_p3*vy_p3+p1y_m3*vy_m3)
//			+coeffs[CY3]*(p1y_p3*tex3D(vel_ref2,ig,jg,y+3)+p1y_m3*tex3D(vel_ref2,ig,jg,y-3))
			+coeffs[CZ4]*(p1s[jl][il+4]*vls[jl][il+4]+p1s[jl][il-4]*vls[jl][il-4])
			+coeffs[CX4]*(p1s[jl+4][il]*vls[jl+4][il]+p1s[jl-4][il]*vls[jl-4][il])
			+coeffs[CY4]*(p1y_p4*vy_p4+p1y_m4*vy_m4)
//			+coeffs[CY4]*(p1y_p4*tex3D(vel_ref2,ig,jg,y+4)+p1y_m4*tex3D(vel_ref2,ig,jg,y-4))
			+p1s[jl][il]+p1s[jl][il]-p0[addr];
		addr+=stride;
	}

}

extern "C" __global__ void wave_kernel_adj_old(float *p0, float *p1, float *p2){
	__shared__ float p1s[BLOCKX_SIZE+2*FAT][BLOCKY_SIZE+2*FAT];

	// define coordinates for the slow axis - everything else will follow from here
	int nblocksz=(n1gpu-2*FAT)/BLOCKZ_SIZE;
	int yblock=(blockIdx.x/nblocksz);
	int y0=FAT+yblock*BLOCKY_SIZE;

	int il=FAT+threadIdx.x; 				// thread-local z coordinate
	int jl=FAT+threadIdx.y;					// thread-local x coordinate

	int ig=(blockIdx.x-yblock*nblocksz)*blockDim.x+il; 	// global z coordinates
	int jg=blockIdx.y*blockDim.y+jl; 	 		// global x coordinates

	float 	p1y_p1, p1y_m1, p1y_p2, p1y_m2, p1y_p3, p1y_m3, p1y_p4, p1y_m4; //here we define the y axis values

	int stride=n1gpu*n2gpu; //this corresponds to distance in elements between wavefield slices
	int addr=ig+n1gpu*jg+stride*y0;
	int addr_fwd=addr-stride-stride-stride-stride; //these (supposedly!) give the memorry location for the current point
		//note that Paulius has a slightly different term here - check one day!

	//copy to shared memory
	p1y_m3=p1[addr_fwd];
	p1y_m2=p1[addr_fwd+=stride];
	p1y_m1=p1[addr_fwd+=stride];
	p1s[jl][il]=p1[addr_fwd+=stride];
	p1y_p1=p1[addr_fwd+=stride];
	p1y_p2=p1[addr_fwd+=stride];
	p1y_p3=p1[addr_fwd+=stride];
	p1y_p4=p1[addr_fwd+=stride];

	// loop over slow axis. ++yl means variable is used before incremented
	for(int yl=0; yl<BLOCKY_SIZE; ++yl){

		int y=y0+yl;

		//update slow axis values
		p1y_m4=p1y_m3;		
		p1y_m3=p1y_m2;		
		p1y_m2=p1y_m1;
		p1y_m1=p1s[jl][il];	
		p1s[jl][il]=p1y_p1;
		p1y_p1=p1y_p2;		
		p1y_p2=p1y_p3;		
		p1y_p3=p1y_p4;
		p1y_p4=p1[addr_fwd+=stride];

    
		if(threadIdx.x<FAT){
			p1s[jl][threadIdx.x]=p1[addr-FAT];
			p1s[jl][il+BLOCKZ_SIZE]=p1[addr+BLOCKZ_SIZE];
		}
		if(threadIdx.y<FAT){
			p1s[threadIdx.y][il]=p1[addr-FAT*n1gpu];
			p1s[jl+BLOCKX_SIZE][il]=p1[addr+BLOCKX_SIZE*n1gpu];
		}
		__syncthreads();

		p2[addr]=
			 coeffs[C0]*p1s[jl][il]*tex3D(vel_ref1,ig,jg,y)
			+coeffs[CZ1]*(p1s[jl][il+1]*tex3D(vel_ref1,ig+1,jg,y)+p1s[jl][il-1]*tex3D(vel_ref1,ig-1,jg,y))
			+coeffs[CX1]*(p1s[jl+1][il]*tex3D(vel_ref1,ig,jg+1,y)+p1s[jl-1][il]*tex3D(vel_ref1,ig,jg-1,y))
			+coeffs[CY1]*(p1y_p1*tex3D(vel_ref1,ig,jg,y+1)+p1y_m1*tex3D(vel_ref1,ig,jg,y-1))
			+coeffs[CZ2]*(p1s[jl][il+2]*tex3D(vel_ref1,ig+2,jg,y)+p1s[jl][il-2]*tex3D(vel_ref1,ig-2,jg,y))
			+coeffs[CX2]*(p1s[jl+2][il]*tex3D(vel_ref1,ig,jg+2,y)+p1s[jl-2][il]*tex3D(vel_ref1,ig,jg-2,y))
			+coeffs[CY2]*(p1y_p2*tex3D(vel_ref1,ig,jg,y+2)+p1y_m2*tex3D(vel_ref1,ig,jg,y-2))
			+coeffs[CZ3]*(p1s[jl][il+3]*tex3D(vel_ref1,ig+3,jg,y)+p1s[jl][il-3]*tex3D(vel_ref1,ig-3,jg,y))
			+coeffs[CX3]*(p1s[jl+3][il]*tex3D(vel_ref1,ig,jg+3,y)+p1s[jl-3][il]*tex3D(vel_ref1,ig,jg-3,y))
			+coeffs[CY3]*(p1y_p3*tex3D(vel_ref1,ig,jg,y+3)+p1y_m3*tex3D(vel_ref1,ig,jg,y-3))
			+coeffs[CZ4]*(p1s[jl][il+4]*tex3D(vel_ref1,ig+4,jg,y)+p1s[jl][il-4]*tex3D(vel_ref1,ig-4,jg,y))
			+coeffs[CX4]*(p1s[jl+4][il]*tex3D(vel_ref1,ig,jg-4,y)+p1s[jl-4][il]*tex3D(vel_ref1,ig,jg+4,y))
			+coeffs[CY4]*(p1y_p4*tex3D(vel_ref1,ig,jg,y+4)+p1y_m4*tex3D(vel_ref1,ig,jg,y-4))
			+p1s[jl][il]+p1s[jl][il]-p0[addr];
			addr+=stride;
	}

}


extern "C" __global__ void wave_kernel_adj2_old(float *p0, float *p1, float *p2){
	__shared__ float p1s[BLOCKX_SIZE+2*FAT][BLOCKY_SIZE+2*FAT];

	// define coordinates for the slow axis - everything else will follow from here
	int nblocksz=(n1gpu-2*FAT)/BLOCKZ_SIZE;
	int yblock=(blockIdx.x/nblocksz);
	int y0=FAT+yblock*BLOCKY_SIZE;

	int il=FAT+threadIdx.x; 				// thread-local z coordinate
	int jl=FAT+threadIdx.y;					// thread-local x coordinate

	int ig=(blockIdx.x-yblock*nblocksz)*blockDim.x+il; 	// global z coordinates
	int jg=blockIdx.y*blockDim.y+jl; 	 		// global x coordinates

	float 	p1y_p1, p1y_m1, p1y_p2, p1y_m2, p1y_p3, p1y_m3, p1y_p4, p1y_m4; //here we define the y axis values

	int stride=n1gpu*n2gpu; //this corresponds to distance in elements between wavefield slices
	int addr=ig+n1gpu*jg+stride*y0;
	int addr_fwd=addr-stride-stride-stride-stride; //these (supposedly!) give the memorry location for the current point
		//note that Paulius has a slightly different term here - check one day!

	//copy to shared memory
	p1y_m3=p1[addr_fwd];
	p1y_m2=p1[addr_fwd+=stride];
	p1y_m1=p1[addr_fwd+=stride];
	p1s[jl][il]=p1[addr_fwd+=stride];
	p1y_p1=p1[addr_fwd+=stride];
	p1y_p2=p1[addr_fwd+=stride];
	p1y_p3=p1[addr_fwd+=stride];
	p1y_p4=p1[addr_fwd+=stride];

	// loop over slow axis. ++yl means variable is used before incremented
	for(int yl=0; yl<BLOCKY_SIZE; ++yl){

		int y=y0+yl;

		//update slow axis values
		p1y_m4=p1y_m3;		
		p1y_m3=p1y_m2;		
		p1y_m2=p1y_m1;
		p1y_m1=p1s[jl][il];	
		p1s[jl][il]=p1y_p1;
		p1y_p1=p1y_p2;		
		p1y_p2=p1y_p3;		
		p1y_p3=p1y_p4;
		p1y_p4=p1[addr_fwd+=stride];

    
		if(threadIdx.x<FAT){
			p1s[jl][threadIdx.x]=p1[addr-FAT];
			p1s[jl][il+BLOCKZ_SIZE]=p1[addr+BLOCKZ_SIZE];
		}
		if(threadIdx.y<FAT){
			p1s[threadIdx.y][il]=p1[addr-FAT*n1gpu];
			p1s[jl+BLOCKX_SIZE][il]=p1[addr+BLOCKX_SIZE*n1gpu];
		}
		__syncthreads();

		p2[addr]=
			 coeffs[C0]*p1s[jl][il]*tex3D(vel_ref2,ig,jg,y)
			+coeffs[CZ1]*(p1s[jl][il+1]*tex3D(vel_ref2,ig+1,jg,y)+p1s[jl][il-1]*tex3D(vel_ref2,ig-1,jg,y))
			+coeffs[CX1]*(p1s[jl+1][il]*tex3D(vel_ref2,ig,jg+1,y)+p1s[jl-1][il]*tex3D(vel_ref2,ig,jg-1,y))
			+coeffs[CY1]*(p1y_p1*tex3D(vel_ref2,ig,jg,y+1)+p1y_m1*tex3D(vel_ref2,ig,jg,y-1))
			+coeffs[CZ2]*(p1s[jl][il+2]*tex3D(vel_ref2,ig+2,jg,y)+p1s[jl][il-2]*tex3D(vel_ref2,ig-2,jg,y))
			+coeffs[CX2]*(p1s[jl+2][il]*tex3D(vel_ref2,ig,jg+2,y)+p1s[jl-2][il]*tex3D(vel_ref2,ig,jg-2,y))
			+coeffs[CY2]*(p1y_p2*tex3D(vel_ref2,ig,jg,y+2)+p1y_m2*tex3D(vel_ref2,ig,jg,y-2))
			+coeffs[CZ3]*(p1s[jl][il+3]*tex3D(vel_ref2,ig+3,jg,y)+p1s[jl][il-3]*tex3D(vel_ref2,ig-3,jg,y))
			+coeffs[CX3]*(p1s[jl+3][il]*tex3D(vel_ref2,ig,jg+3,y)+p1s[jl-3][il]*tex3D(vel_ref2,ig,jg-3,y))
			+coeffs[CY3]*(p1y_p3*tex3D(vel_ref2,ig,jg,y+3)+p1y_m3*tex3D(vel_ref2,ig,jg,y-3))
			+coeffs[CZ4]*(p1s[jl][il+4]*tex3D(vel_ref2,ig+4,jg,y)+p1s[jl][il-4]*tex3D(vel_ref2,ig-4,jg,y))
			+coeffs[CX4]*(p1s[jl+4][il]*tex3D(vel_ref2,ig,jg-4,y)+p1s[jl-4][il]*tex3D(vel_ref2,ig,jg+4,y))
			+coeffs[CY4]*(p1y_p4*tex3D(vel_ref2,ig,jg,y+4)+p1y_m4*tex3D(vel_ref2,ig,jg,y-4))
			+p1s[jl][il]+p1s[jl][il]-p0[addr];
			addr+=stride;
	}

}

extern "C" __global__ void wave_kernelx(float *p0, float *p1, float *p2){
	__shared__ float p1s[BLOCKX_SIZE+2*FAT][BLOCKY_SIZE+2*FAT];

	// define coordinates for the slow axis - everything else will follow from here
	int nblocksx=(n2gpu-2*FAT)/BLOCKX_SIZE;
	int yblock=(blockIdx.y/nblocksx);
	int y0=FAT+yblock*BLOCKY_SIZE;

	int il=FAT+threadIdx.x; 				// thread-local z coordinate
	int jl=FAT+threadIdx.y;					// thread-local x coordinate

	int jg=(blockIdx.y-yblock*nblocksx)*blockDim.y+jl; 	// global z coordinates
	int ig=blockIdx.x*blockDim.x+il; 	 		// global x coordinates

	float 	p1y_p1, p1y_m1, p1y_p2, p1y_m2, p1y_p3, p1y_m3, p1y_p4, p1y_m4; //here we define the y axis values

	int stride=n1gpu*n2gpu; //this corresponds to distance in elements between wavefield slices
	int addr=ig+n1gpu*jg+stride*y0;
	int addr_fwd=addr-stride-stride-stride-stride; //these (supposedly!) give the memorry location for the current point
		//note that Paulius has a slightly different term here - check one day!

	//copy to shared memory
	p1y_m3=p1[addr_fwd];
	p1y_m2=p1[addr_fwd+=stride];
	p1y_m1=p1[addr_fwd+=stride];
	p1s[jl][il]=p1[addr_fwd+=stride];
	p1y_p1=p1[addr_fwd+=stride];
	p1y_p2=p1[addr_fwd+=stride];
	p1y_p3=p1[addr_fwd+=stride];
	p1y_p4=p1[addr_fwd+=stride];

	// loop over slow axis. ++yl means variable is used before incremented
	for(int yl=0; yl<BLOCKY_SIZE; ++yl){

		int y=y0+yl;

		//update slow axis values
		p1y_m4=p1y_m3;		
		p1y_m3=p1y_m2;		
		p1y_m2=p1y_m1;
		p1y_m1=p1s[jl][il];	
		p1s[jl][il]=p1y_p1;
		p1y_p1=p1y_p2;		
		p1y_p2=p1y_p3;		
		p1y_p3=p1y_p4;
		p1y_p4=p1[addr_fwd+=stride];

    
		if(threadIdx.x<FAT){
			p1s[jl][threadIdx.x]=p1[addr-FAT];
			p1s[jl][il+BLOCKZ_SIZE]=p1[addr+BLOCKZ_SIZE];
		}
		if(threadIdx.y<FAT){
			p1s[threadIdx.y][il]=p1[addr-FAT*n1gpu];
			p1s[jl+BLOCKX_SIZE][il]=p1[addr+BLOCKX_SIZE*n1gpu];
		}
		__syncthreads();

		p2[addr]=//10;
		tex3D(vel_ref1,ig,jg,y)*
			(
			 coeffs[C0]*p1s[jl][il]
			+coeffs[CZ1]*(p1s[jl][il+1]+p1s[jl][il-1])
			+coeffs[CX1]*(p1s[jl+1][il]+p1s[jl-1][il])
			+coeffs[CY1]*(p1y_p1+p1y_m1)
			+coeffs[CZ2]*(p1s[jl][il+2]+p1s[jl][il-2])
			+coeffs[CX2]*(p1s[jl+2][il]+p1s[jl-2][il])
			+coeffs[CY2]*(p1y_p2+p1y_m2)
			+coeffs[CZ3]*(p1s[jl][il+3]+p1s[jl][il-3])
			+coeffs[CX3]*(p1s[jl+3][il]+p1s[jl-3][il])
			+coeffs[CY3]*(p1y_p3+p1y_m3)
			+coeffs[CZ4]*(p1s[jl][il+4]+p1s[jl][il-4])
			+coeffs[CX4]*(p1s[jl+4][il]+p1s[jl-4][il])
			+coeffs[CY4]*(p1y_p4+p1y_m4)
			)
			+p1s[jl][il]+p1s[jl][il]-p0[addr];
			addr+=stride;
	}
}

extern "C" __global__ void wave_kernelx2(float *p0, float *p1, float *p2){
	__shared__ float p1s[BLOCKX_SIZE+2*FAT][BLOCKY_SIZE+2*FAT];

	// define coordinates for the slow axis - everything else will follow from here
	int nblocksx=(n2gpu-2*FAT)/BLOCKX_SIZE;
	int yblock=(blockIdx.y/nblocksx);
	int y0=FAT+yblock*BLOCKY_SIZE;

	int il=FAT+threadIdx.x; 				// thread-local z coordinate
	int jl=FAT+threadIdx.y;					// thread-local x coordinate

	int jg=(blockIdx.y-yblock*nblocksx)*blockDim.y+jl; 	// global z coordinates
	int ig=blockIdx.x*blockDim.x+il; 	 		// global x coordinates

	float 	p1y_p1, p1y_m1, p1y_p2, p1y_m2, p1y_p3, p1y_m3, p1y_p4, p1y_m4; //here we define the y axis values

	int stride=n1gpu*n2gpu; //this corresponds to distance in elements between wavefield slices
	int addr=ig+n1gpu*jg+stride*y0;
	int addr_fwd=addr-stride-stride-stride-stride; //these (supposedly!) give the memorry location for the current point
		//note that Paulius has a slightly different term here - check one day!

	//copy to shared memory
	p1y_m3=p1[addr_fwd];
	p1y_m2=p1[addr_fwd+=stride];
	p1y_m1=p1[addr_fwd+=stride];
	p1s[jl][il]=p1[addr_fwd+=stride];
	p1y_p1=p1[addr_fwd+=stride];
	p1y_p2=p1[addr_fwd+=stride];
	p1y_p3=p1[addr_fwd+=stride];
	p1y_p4=p1[addr_fwd+=stride];

	// loop over slow axis. ++yl means variable is used before incremented
	for(int yl=0; yl<BLOCKY_SIZE; ++yl){

		int y=y0+yl;

		//update slow axis values
		p1y_m4=p1y_m3;		
		p1y_m3=p1y_m2;		
		p1y_m2=p1y_m1;
		p1y_m1=p1s[jl][il];	
		p1s[jl][il]=p1y_p1;
		p1y_p1=p1y_p2;		
		p1y_p2=p1y_p3;		
		p1y_p3=p1y_p4;
		p1y_p4=p1[addr_fwd+=stride];

    
		if(threadIdx.x<FAT){
			p1s[jl][threadIdx.x]=p1[addr-FAT];
			p1s[jl][il+BLOCKZ_SIZE]=p1[addr+BLOCKZ_SIZE];
		}
		if(threadIdx.y<FAT){
			p1s[threadIdx.y][il]=p1[addr-FAT*n1gpu];
			p1s[jl+BLOCKX_SIZE][il]=p1[addr+BLOCKX_SIZE*n1gpu];
		}
		__syncthreads();

		p2[addr]=//10;
		tex3D(vel_ref2,ig,jg,y)*
			(
			 coeffs[C0]*p1s[jl][il]
			+coeffs[CZ1]*(p1s[jl][il+1]+p1s[jl][il-1])
			+coeffs[CX1]*(p1s[jl+1][il]+p1s[jl-1][il])
			+coeffs[CY1]*(p1y_p1+p1y_m1)
			+coeffs[CZ2]*(p1s[jl][il+2]+p1s[jl][il-2])
			+coeffs[CX2]*(p1s[jl+2][il]+p1s[jl-2][il])
			+coeffs[CY2]*(p1y_p2+p1y_m2)
			+coeffs[CZ3]*(p1s[jl][il+3]+p1s[jl][il-3])
			+coeffs[CX3]*(p1s[jl+3][il]+p1s[jl-3][il])
			+coeffs[CY3]*(p1y_p3+p1y_m3)
			+coeffs[CZ4]*(p1s[jl][il+4]+p1s[jl][il-4])
			+coeffs[CX4]*(p1s[jl+4][il]+p1s[jl-4][il])
			+coeffs[CY4]*(p1y_p4+p1y_m4)
			)
			+p1s[jl][il]+p1s[jl][il]-p0[addr];
			addr+=stride;
	}
}

